@@ -5,68 +5,35 @@ use base 'Curses::UI::Window';
 use Curses qw/KEY_ENTER/;
 use Socialtext::Wikrad qw/$App/;
 use Socialtext::Resting;
+use JSON;
+use Data::Dumper;
 
 sub new {
     my $class = shift;
     my $self  = $class->SUPER::new(@_);
 
-    #######################################
-    # Create the Workspace label and field
-    #######################################
-    my $wksp_cb = sub { toggle_editable( shift, \&workspace_change ) };
-    my $w = $self->{wksp} = $self->add_field('Workspace:', $wksp_cb,
-        -text => $App->{rester}->workspace,
-        -width => 18,
-        -x => 1,
-    );
-
-    #######################################
-    # Create the Page label and field
-    #######################################
-    my $page_cb = sub { toggle_editable( shift, sub { $App->load_page } ) };
-    my $p = $self->{page_box} = $self->add_field('Page:', $page_cb,
-        -width => 45,
-        -x => 32,
-    );
-
-    #######################################
-    # Create the Tag label and field
-    #######################################
-    my $tag_cb = sub { toggle_editable( shift, \&tag_change ) };
-    my $t = $self->{tag} = $self->add_field('Tag:', $tag_cb,
-        -width => 15,
-        -x => 85,
-    );
-
-    $self->add(undef, 'Label',
-        -x => 107,
-        -bold => 1,
-        -text => "Help: hit '?'"
-    );
-
-    #######################################
-    # Create the page Viewer
-    #######################################
-    my $v = $self->{viewer} = $self->add(
-        'viewer', 'Socialtext::Wikrad::PageViewer',
-        -border => 1,
-        -y      => 1,
-    );
+    $self->_create_ui_widgets;
 
+    my ($v, $p, $w, $t) = map { $self->{$_} } 
+                          qw/viewer page_box workspace_box tag_box/;
     $v->focus;
-    $v->set_binding( \&choose_frontlink, 'g' );
-    $v->set_binding( \&choose_backlink,  'B' );
-    $v->set_binding( \&show_help,        '?' );
-    $v->set_binding( \&recently_changed, 'r' );
-    $v->set_binding( \&show_uri,         'u' );
-    $v->set_binding( \&show_includes,    'i' );
+    $v->set_binding( \&choose_frontlink,         'g' );
+    $v->set_binding( \&choose_backlink,          'B' );
+    $v->set_binding( \&show_help,                '?' );
+    $v->set_binding( \&recently_changed,         'r' );
+    $v->set_binding( \&show_uri,                 'u' );
+    $v->set_binding( \&show_includes,            'i' );
+    $v->set_binding( \&clone_page,               'c' );
+    $v->set_binding( \&clone_page_from_template, 'C' );
+    $v->set_binding( \&show_metadata,            'm' );
+    $v->set_binding( \&add_pagetag,              'T' );
 
     $v->set_binding( sub { editor() },                  'e' );
     $v->set_binding( sub { editor('--pull-includes') }, 'E' );
     $v->set_binding( sub { $v->focus },                 'v' );
-    $v->set_binding( sub { $p->focus; $page_cb->($p) }, 'p');
-    $v->set_binding( sub { $w->focus; $wksp_cb->($w) }, 'w');
-    $v->set_binding( sub { $t->focus; $tag_cb->($t) },  't');
+    $v->set_binding( sub { $p->focus; $self->{cb}{page}->($p) },      'p' );
+    $v->set_binding( sub { $w->focus; $self->{cb}{workspace}->($w) }, 'w' );
+    $v->set_binding( sub { $t->focus; $self->{cb}{tag}->($t) },       't' );
 
     $v->set_binding( sub { $v->viewer_enter }, KEY_ENTER );
     $v->set_binding( sub { $App->go_back }, 'b' );
@@ -103,6 +70,9 @@ Navigation:
  b - go back
  u - show the uri for the current page
  i - show included pages
+ c - clone this page
+ m - show page metadata (tags, revision)
+ T - Tag page
 
 Movement:
  ENTER   - jump to page [under cursor]
@@ -120,32 +90,45 @@ Ctrl-q / Ctrl-c / q - quit
 EOT
 }
 
-sub listbox {
-    my $self = shift;
-    $App->{win}->add('listbox', 'Socialtext::Wikrad::Listbox', @_)->focus;
+sub add_pagetag {
+    my $r = $App->{rester};
+    $App->{cui}->status('Fetching page tags ...');
+    $r->accept('text/plain');
+    my $page_name = $App->get_page;
+    my @tags = $r->get_pagetags($page_name);
+    $App->{cui}->nostatus;
+    my $question = "Enter new tags, separate with commas, prefix with '-' to remove\n  ";
+    if (@tags) {
+        $question .= join(", ", @tags) . "\n";
+    }
+    my $newtags = $App->{cui}->question($question) || '';
+    my @new_tags = split(/\s*,\s*/, $newtags);
+    if (@new_tags) {
+        $App->{cui}->status("Tagging $page_name ...");
+        for my $t (@new_tags) {
+            if ($t =~ s/^-//) {
+                eval { $r->delete_pagetag($page_name, $t) };
+            }
+            else {
+                $r->put_pagetag($page_name, $t);
+            }
+        }
+        $App->{cui}->nostatus;
+    }
 }
 
-sub add_field {
-    my $self = shift;
-    my $desc = shift;
-    my $cb = shift;
-    my %args = @_;
-    my $x = $args{-x} || 0;
-
-    $self->add(undef, 'Label',
-        -bold => 1,
-        -text => $desc,
-        -x => $x,
+sub show_metadata {
+    my $r = $App->{rester};
+    $App->{cui}->status('Fetching page metadata ...');
+    $r->accept('application/json');
+    my $page_name = $App->get_page;
+    my $json_text = $r->get_page($page_name);
+    my $page_data = jsonToObj($json_text);
+    $App->{cui}->nostatus;
+    $App->{cui}->dialog(
+        -title => "$page_name metadata",
+        -message => Dumper $page_data,
     );
-    $args{-x} = $x + length($desc) + 1;
-    my $w = $self->add(undef, 'TextEntry', 
-        -singleline => 1,
-        -sbborder => 1,
-        -readonly => 1,
-        %args,
-    );
-    $w->set_binding( sub { $cb->($w) }, KEY_ENTER );
-    return $w;
 }
 
 sub show_uri {
@@ -153,7 +136,37 @@ sub show_uri {
     my $uri = $r->server . '/' . $r->workspace 
               . '/index.cgi?' 
               . Socialtext::Resting::_name_to_id($App->get_page);
-    $App->{cui}->dialog( -title => "Current page:", -message => $uri );
+    $App->{cui}->dialog( -title => "Current page:", -message => " $uri" );
+}
+
+sub clone_page {
+    my @args = @_; # obj, key, args
+    my $template_page = $args[2] || $App->get_page;
+    my $r = $App->{rester};
+    $r->accept('text/x.socialtext-wiki');
+    my $template = $r->get_page($template_page);
+    my $new_page = $App->{cui}->question("Title for new page:");
+    $App->{cui}->status("Creating page ...");
+    $r->put_page($new_page, $template);
+    $r->accept('text/plain');
+    my @tags = grep { $_ ne 'template' } $r->get_pagetags($template_page);
+    $r->put_pagetag($new_page, $_) for @tags;
+    $App->{cui}->nostatus;
+
+    $App->set_page($new_page);
+}
+
+sub clone_page_from_template {
+    my $tag = 'template';
+    $App->{cui}->status('Fetching pages tagged $tag...');
+    $App->{rester}->accept('text/plain');
+    my @pages = $App->{rester}->get_taggedpages($tag);
+    $App->{cui}->nostatus;
+    $App->{win}->listbox(
+        -title => 'Choose a template',
+        -values => \@pages,
+        change_cb => sub { clone_page(undef, undef, shift) },
+    );
 }
 
 sub show_includes {
@@ -163,9 +176,11 @@ sub show_includes {
     my $page_text = $viewer->text;
     while($page_text =~ m/\{include:? \[(.+?)\]\}/g) {
         my $included_page = $1;
+        $r->accept('text/x.socialtext-wiki');
         my $included_text = $r->get_page($included_page);
         my $new_text = "-----Included Page----- [$included_page]\n"
-                       . "$included_text\n";
+                       . "$included_text\n"
+                       . "-----End Include----- \n";
         $page_text =~ s/{include:? \[\Q$included_page\E\]}/$new_text/;
     }
     $viewer->text($page_text);
@@ -175,6 +190,7 @@ sub show_includes {
 sub recently_changed {
     my $r = $App->{rester};
     $App->{cui}->status('Fetching recent changes ...');
+    $r->accept('text/plain');
     my @recent = $r->get_taggedpages('Recent changes');
     $App->{cui}->nostatus;
     $App->{win}->listbox(
@@ -201,6 +217,7 @@ sub choose_link {
     my $arg = shift;
     my $page = $App->get_page;
     $App->{cui}->status("Fetching ${text}s");
+    $App->{rester}->accept('text/plain');
     my @links = $App->{rester}->$method($page, $arg);
     $App->{cui}->nostatus;
     if (@links) {
@@ -219,26 +236,27 @@ sub choose_link {
 }
 
 sub editor {
-    my $extra_args = shift || '';
+    my @extra_args = @_;
     $App->{cui}->status('Editing page');
     $App->{cui}->leave_curses;
     my $r = $App->{rester};
     my $workspace = $r->workspace;
     my $page = $App->get_page;
     my $server = $r->server;
-    system("wikedit -s '$server' -w '$workspace' $extra_args '$page'");
+    system("wikedit", '-s', $server, '-w', $workspace, @extra_args, $page);
     $App->{cui}->reset_curses;
     $App->load_page;
 }
 
 sub workspace_change {
-    my $new_wksp = $App->{win}{wksp}->text;
+    my $new_wksp = $App->{win}{workspace_box}->text;
     my $r = $App->{rester};
     if ($new_wksp) {
         $App->set_page(undef, $new_wksp);
     }
     else {
         $App->{cui}->status('Fetching list of workspaces ...');
+        $r->accept('text/plain');
         my @workspaces = $r->get_workspaces;
         $App->{cui}->nostatus;
         $App->{win}->listbox(
@@ -254,11 +272,18 @@ sub workspace_change {
 
 sub tag_change {
     my $r = $App->{rester};
-    my $tag = $App->{win}{tag}->text;
+    my $tag = $App->{win}{tag_box}->text;
 
     my $chose_tagged_page = sub {
         my $tag = shift;
+        $App->{cui}->status('Fetching tagged pages ...');
+        $r->accept('text/plain');
         my @pages = $r->get_taggedpages($tag);
+        $App->{cui}->nostatus;
+        if (@pages == 0) {
+            $App->{cui}->dialog("No pages tagged '$tag' found ...");
+            return;
+        }
         $App->{win}->listbox(
             -title => 'Choose a tagged page',
             -values => \@pages,
@@ -272,7 +297,10 @@ sub tag_change {
         $chose_tagged_page->($tag);
     }
     else {
+        $App->{cui}->status('Fetching workspace tags ...');
+        $r->accept('text/plain');
         my @tags = $r->get_workspace_tags;
+        $App->{cui}->nostatus;
         $App->{win}->listbox(
             -title => 'Choose a tag:',
             -values => \@tags,
@@ -306,4 +334,81 @@ sub toggle_editable {
     $w->set_binding( sub { toggle_editable($w, $cb) }, KEY_ENTER );
 }
 
+sub _create_ui_widgets {
+    my $self = shift;
+    #######################################
+    # Create the Workspace label and field
+    #######################################
+    my $wksp_cb = sub { toggle_editable( shift, \&workspace_change ) };
+    $self->{cb}{workspace} = $wksp_cb;
+    $self->{workspace_box} = $self->add_field('Workspace:', $wksp_cb,
+        -text => $App->{rester}->workspace,
+        -width => 18,
+        -x => 1,
+    );
+
+    #######################################
+    # Create the Page label and field
+    #######################################
+    my $page_cb = sub { toggle_editable( shift, sub { $App->load_page } ) };
+    $self->{cb}{page} = $page_cb;
+    $self->{page_box} = $self->add_field('Page:', $page_cb,
+        -width => 45,
+        -x => 32,
+    );
+
+    #######################################
+    # Create the Tag label and field
+    #######################################
+    my $tag_cb = sub { toggle_editable( shift, \&tag_change ) };
+    $self->{cb}{tag} = $tag_cb;
+    $self->{tag_box} = $self->add_field('Tag:', $tag_cb,
+        -width => 15,
+        -x => 85,
+    );
+
+    $self->add(undef, 'Label',
+        -x => 107,
+        -bold => 1,
+        -text => "Help: hit '?'"
+    );
+
+    #######################################
+    # Create the page Viewer
+    #######################################
+    $self->{viewer} = $self->add(
+        'viewer', 'Socialtext::Wikrad::PageViewer',
+        -border => 1,
+        -y      => 1,
+    );
+}
+
+sub listbox {
+    my $self = shift;
+    $App->{win}->add('listbox', 'Socialtext::Wikrad::Listbox', @_)->focus;
+}
+
+sub add_field {
+    my $self = shift;
+    my $desc = shift;
+    my $cb = shift;
+    my %args = @_;
+    my $x = $args{-x} || 0;
+
+    $self->add(undef, 'Label',
+        -bold => 1,
+        -text => $desc,
+        -x => $x,
+    );
+    $args{-x} = $x + length($desc) + 1;
+    my $w = $self->add(undef, 'TextEntry', 
+        -singleline => 1,
+        -sbborder => 1,
+        -readonly => 1,
+        %args,
+    );
+    $w->set_binding( sub { $cb->($w) }, KEY_ENTER );
+    return $w;
+}
+
 1;
